import { ParserOptions } from '@/options';
import { RawSourceMap } from 'source-map';

declare type arithmeticOperator = '+' | '-' | '*' | '/' | '%';

declare interface AssignStatement extends Node_2 {
    type: NodeTypes.ASSIGN;
    left: VariableNode | TextNode | VarKeyNode;
    right: ListNode | TextNode;
}

declare interface BinaryNode extends Node_2 {
    type: NodeTypes.BINARY;
    operator: OperatorNode;
    left: TextNode | VariableNode | BinaryNode;
    right: TextNode | VariableNode | BinaryNode;
}

declare interface BodyStatement extends Node_2 {
    type: NodeTypes.BODY;
    children: Statement[];
}

declare interface ChildStatement extends Node_2 {
    type: NodeTypes.CHILD;
    selector: TextNode | PlaceholderNode | ListNode;
    children: (Statement | CodegenNode)[];
}

declare type CodegenNode = TextNode | ProgCodeGenNode;

declare interface CodegenOptions {
    /**
     * Generate source map?
     * @default false
     */
    sourceMap?: boolean;
}

declare interface CodegenResult {
    code: string;
    ast: RootNode;
    map?: RawSourceMap;
}

declare type comparisonOperator = '>' | '<' | '>=' | '<=' | '==' | '!=';

export declare function compile(scss: string, options?: CompilerOptions): CodegenResult;

declare interface CompilerError extends SyntaxError {
    code: ErrorCodes;
    loc?: SourceLocation;
}

declare type CompilerOptions = ParserOptions_2 & TransformOptions & CodegenOptions;

declare interface EachStatement extends Node_2 {
    type: NodeTypes.EACHSTATEMENT;
    left: VariableNode;
    right: VariableNode;
    body: ChildStatement;
}

declare interface EmptyNode extends Node_2 {
    type: NodeTypes.EMPTY;
}

declare class Environment {
    vars: {
        [name: string]: any;
    };
    parent: Environment | null;
    constructor(parent: Environment | null);
    extend(): Environment;
    get(name: string): any;
    def(name: string, value: any): any;
}

declare const enum ErrorCodes {
    UNKNONWN_TOKEN_TYPE = 0,
    INVALID_LOC_POSITION = 1,
    EXPECT_TEXT_NODE_AFTER_OPERATOR_NODE = 2
}

declare interface ErrorStatement extends Node_2 {
    type: NodeTypes.ERROR;
    value: ListNode;
}

declare interface ExtendStatement extends Node_2 {
    type: NodeTypes.EXTEND;
    param: TextNode | PlaceholderNode;
}

declare type FileSourceMap = {
    [key: string]: string;
};

declare interface IdentifierNode extends Node_2 {
    type: NodeTypes.IDENTIFIER;
    value: string;
}

declare interface IfStatement extends Node_2 {
    type: NodeTypes.IFSTATEMENT;
    test: BinaryNode | TextNode;
    consequent: BodyStatement;
    alternate: IfStatement | BodyStatement | null;
}

declare interface ImportStatement extends Node_2 {
    type: NodeTypes.IMPORT;
    params: TextNode[];
}

declare interface IncludeStatement extends Node_2 {
    type: NodeTypes.INCLUDE;
    id: IdentifierNode;
    args: (TextNode | VariableNode | BinaryNode | AssignStatement)[];
}

declare interface ListNode extends Node_2 {
    type: NodeTypes.LIST;
    value: SimpleExpressionNode[];
}

declare interface MixinStatement extends Node_2 {
    type: NodeTypes.MIXIN;
    id: IdentifierNode;
    params: (VariableNode | AssignStatement)[];
    body: BodyStatement;
}

declare interface Node_2 {
    [key: string]: any;
    type: NodeTypes;
    loc: SourceLocation;
}

declare type NodeTransform = (node: Node_2, context: TransformContext) => void | (() => void) | (() => void)[];

declare const enum NodeTypes {
    TEXT = "TEXT",
    VARIABLE = "VARIABLE",
    PUNC = "PUNC",
    OPERATOR = "OPERATOR",
    VAR_KEY = "VAR_KEY",
    PLACEHOLDER = "PLACEHOLDER",
    KEYWORD = "KEYWORD",
    IDENTIFIER = "IDENTIFIER",
    EMPTY = "EMPTY",
    /**
     * https://sass-lang.com/documentation/values/lists
     * any expressions separated with spaces or commas count as a list;
     * iterable eg: $each value in list
    */
    LIST = "LIST",
    BINARY = "BINARY",
    /**
     * Statement
     * */
    BODY = "BODY",
    CHILD = "CHILD",
    ASSIGN = "ASSIGN",
    IMPORT = "IMPORT",
    INCLUDE = "INCLUDE",
    EXTEND = "EXTEND",
    MIXIN = "MIXIN",
    ERROR = "ERROR",
    IFSTATEMENT = "IFSTATEMENT",
    EACHSTATEMENT = "EACHSTATEMENT",
    RootNode = "RootNode"
}

declare interface OperatorNode extends Node_2 {
    type: NodeTypes.OPERATOR;
    value: arithmeticOperator | comparisonOperator;
}

export declare function parse(scss: string, options: ParserOptions): RootNode;

declare interface ParserOptions_2 {
    /**
    * Filename for source map generation.
    */
    filename: string;
    source: string;
}

declare interface PlaceholderNode extends Node_2 {
    type: NodeTypes.PLACEHOLDER;
    value: string;
}

declare interface Position_2 {
    offset: number;
    line: number;
    column: number;
}

declare type ProgCodeGenNode = AssignStatement | ChildStatement | EmptyNode;

declare interface PuncNode extends Node_2 {
    type: NodeTypes.PUNC;
    value: puncType;
}

declare type puncType = '(' | ')' | ',' | ';' | '#';

declare interface RootNode extends Node_2 {
    type: NodeTypes.RootNode;
    fileSourceMap: FileSourceMap;
    children: (Statement | ProgCodeGenNode)[];
}

export declare function run(sourceDir: string, outputDir?: string, options?: RunOptions): void;

export declare interface RunOptions {
    genOtherInfo: boolean;
    sourceMap: boolean;
}

declare type SimpleExpressionNode = TextNode | PuncNode | OperatorNode | VariableNode | VarKeyNode | BinaryNode | ListNode;

declare interface SourceLocation {
    filename: string;
    start: Position_2;
    end: Position_2;
}

declare type Statement = BodyStatement | ChildStatement | AssignStatement | ImportStatement | IncludeStatement | ExtendStatement | MixinStatement | ErrorStatement | IfStatement | EachStatement;

declare interface TextNode extends Node_2 {
    type: NodeTypes.TEXT;
    value: string;
}

declare interface TransformContext extends Required<TransformOptions> {
    root: RootNode;
    currentNode: Node_2 | null;
    parent: Node_2 | null;
    childIndex: number;
    env: Environment;
    replaceNode(node?: Node_2): void;
    onNodeRemoved(): void;
}

declare interface TransformOptions {
    /**
     * An array of node trasnforms to be applied to every AST node.
     */
    nodeTransforms?: NodeTransform[];
    sourceDir?: string;
    onError?: (error: CompilerError) => void;
}

declare interface VariableNode extends Node_2 {
    type: NodeTypes.VARIABLE;
    value: string;
}

declare interface VarKeyNode extends Node_2 {
    type: NodeTypes.VAR_KEY;
    value: string;
}

export { }
