import fs from 'fs';
import path from 'path';
import require$$0 from 'util';

function input_stream(input) {
    let offset = 0, line = 1, column = 1;
    return {
        next,
        peek,
        setCoordination,
        getCoordination,
        eof,
        croak,
    };
    function next() {
        let ch = input.charAt(offset++);
        if (ch == "\n")
            line++, column = 1;
        else
            column++;
        return ch;
    }
    function setCoordination(coordination) {
        offset = coordination.offset;
        line = coordination.line;
        column = coordination.column;
    }
    function getCoordination() {
        return {
            offset,
            line,
            column
        };
    }
    function peek() {
        return input.charAt(offset);
    }
    function eof() {
        return peek() == "";
    }
    function croak(msg) {
        throw new Error(msg + " (" + line + ":" + column + ")");
    }
}

const locStub = {
    filename: '',
    start: { line: 1, column: 1, offset: 0 },
    end: { line: 1, column: 1, offset: 0 }
};
function createAssignStatement(left, right) {
    return {
        type: "ASSIGN" /* ASSIGN */,
        left: left,
        right: right,
        loc: locStub
    };
}
function createListNode(list) {
    return {
        type: "LIST" /* LIST */,
        value: list,
        loc: {
            start: list[0].loc.start,
            end: list[list.length - 1].loc.end,
            filename: list[0].loc.filename
        }
    };
}

const debug = (function () {
    let isDebug = false, count = 0;
    return () => {
        if (count++ > 20 && isDebug) {
            return true;
        }
        return false;
    };
})();
function is_calculate_op_char(ch) {
    let op_chars = ' + - * / % ';
    return op_chars.indexOf(' ' + ch + ' ') >= 0;
}
function is_punc(ch) {
    return ",;(){}#".indexOf(ch) >= 0; // support expr { #{var}:var }
}
const PRECEDENCE = {
    "=": 1,
    "||": 2,
    "&&": 3,
    "<": 7, ">": 7, "<=": 7, ">=": 7, "==": 7, "!=": 7,
    "+": 10, "-": 10,
    "*": 20, "/": 20, "%": 20,
};
/**
 * fill whitespace between tokens which is removed in lexical analyze
 *
 */
function fillWhitespace(tokens) {
    if (tokens.length <= 1)
        return tokens;
    let list = [], whitespaceToken = addNodeEmptyLocation({
        type: "TEXT" /* TEXT */,
        value: ' '
    }), curIndex = 0, curToken, nextToken;
    while (curIndex < tokens.length - 1) {
        curToken = tokens[curIndex];
        nextToken = tokens[curIndex + 1];
        list.push(curToken);
        if (nextToken.loc.start.offset > curToken.loc.end.offset) {
            list.push(whitespaceToken);
        }
        curIndex++;
    }
    list.push(tokens[curIndex]);
    return list;
}
class Environment {
    constructor(parent) {
        this.vars = Object.create(parent ? parent.vars : null);
        this.parent = parent;
    }
    extend() {
        return new Environment(this);
    }
    // lookup: function (name) {
    //     let scope = this;
    //     while (scope) {
    //         if (Object.prototype.hasOwnProperty.call(scope.vars, name))
    //             return scope;
    //         scope = scope.parent;
    //     }
    // },
    get(name) {
        if (name in this.vars)
            return this.vars[name];
        throw new Error("Undefined variable " + name);
    }
    // set: function (name, value) {
    //     let scope = this.lookup(name);
    //     // let's not allow defining globals from a nested environment
    //     if (!scope && this.parent)
    //         throw new Error("Undefined variable " + name);
    //     return (scope || this).vars[name] = value;
    // },
    def(name, value) {
        return this.vars[name] = value;
    }
}
/**
 * todos: optimize deep_clone
 */
function deepClone(obj) {
    return JSON.parse(JSON.stringify(obj));
}
// Some expressions, e.g. sequence and conditional expressions, are never
// associated with generated code, so their source locations are just a empty.
function addNodeEmptyLocation(node) {
    node.loc = locStub;
    return node;
}
function createEmptyNode() {
    return addNodeEmptyLocation({
        type: "EMPTY" /* EMPTY */,
    });
}
function isEmptyNode(node) {
    return node.type === "EMPTY" /* EMPTY */;
}
// function createPromiseCallback() {
//     var resolve, reject;
//     var promise = new Promise(function (_resolve, _reject) {
//         resolve = _resolve;
//         reject = _reject;
//     });
//     var cb = function (err, res) {
//         if (err) { return reject(err) }
//         resolve(res || '');
//     };
//     return { promise: promise, cb: cb }
// }
// advance by mutation without cloning (for performance reasons), since this
// gets called a lot in the parser
function advancePositionWithMutation(pos, source, numberOfCharacters = source.length) {
    let linesCount = 0;
    let lastNewLinePos = -1;
    for (let i = 0; i < numberOfCharacters; i++) {
        if (source.charCodeAt(i) === 10 /* newline char code */) {
            linesCount++;
            lastNewLinePos = i;
        }
    }
    pos.offset += numberOfCharacters;
    pos.line += linesCount;
    pos.column =
        lastNewLinePos === -1
            ? pos.column + numberOfCharacters
            : numberOfCharacters - lastNewLinePos;
    return pos;
}

/**
 *
 * @param {input_stream processed stream method} input
 * { type: NodeTypes.PUNC, value: "(" }           // punctuation: parens((|)), comma(,), semicolon(;) etc.
 * { type: NodeTypes.TEXT, value: "12px" }
 * { type: NodeTypes.VARIABLE, value: "$height" }      // identifiers
 * { type: NodeTypes.KEYWORD, value: NodeTypes.INCLUDE}    // keywords below
 * { type: NodeTypes.PLACEHOLDER, value: "%str" }      //  % started string contains op char '%'
 * { type: NodeTypes.OPERATOR, value: "!=" }            // + - % * / != ==
 */
function lex(input) {
    // let current = null;
    let keywords = ' @extend @mixin @include @import @if @else @error @each ', comparison_op_chars = '!=><', comparison_op_tokens = ['==', '!=', '>=', '<=', '>', '<'];
    return {
        next,
        peek,
        eliminateWhitespace,
        eof,
        getCoordination: input.getCoordination,
        setCoordination: input.setCoordination,
        croak: input.croak
    };
    function is_comment_char(ch) {
        return "/".indexOf(ch) >= 0;
    }
    function is_id_start(ch) {
        return /[$]/.test(ch);
    }
    function skip_comment() {
        read_while(function (ch) { return ch != "\n"; });
        input.next();
    }
    function is_comparison_op_char(ch) {
        // return "+-*/%=&|<>!".indexOf(ch) >= 0;
        return comparison_op_chars.indexOf(ch) >= 0;
    }
    function is_comparison_op_tokens(str) {
        return comparison_op_tokens.indexOf(str) >= 0;
    }
    function is_assign_char(ch) {
        return ":".indexOf(ch) >= 0;
    }
    function is_base_char(ch) {
        return /[a-z0-9_\.\#\@\%\-"'&\[\]]/i.test(ch);
        // return !is_punc(ch);
    }
    function is_id_char_limit(ch) {
        return is_id_start(ch) || /[a-z0-9_-]/i.test(ch); // sass变量名限制
    }
    function is_keyword(kw) {
        return keywords.indexOf(" " + kw + " ") >= 0;
    }
    function is_keyword_start(ch) {
        return ch === '@';
    }
    function is_placeholder_start(ch) {
        return ch === '%';
    }
    function is_whitespace(ch) {
        return " \t\n".indexOf(ch) >= 0;
    }
    function read_while(predicate) {
        let str = "";
        while (!input.eof() && predicate(input.peek()))
            str += input.next();
        return str;
    }
    function read_keyword() {
        let kw = read_while(is_base_char);
        if (!is_keyword(kw)) {
            return input.croak(`Unknown keyword ${kw}`);
        }
        return {
            type: "KEYWORD" /* KEYWORD */,
            value: kw
        };
    }
    function read_assign_char() {
        return {
            type: "ASSIGN" /* ASSIGN */,
            value: input.next()
        };
    }
    function read_ident() {
        let id = read_while(is_id_char_limit);
        return {
            type: "VARIABLE" /* VARIABLE */,
            value: id.trim()
        };
    }
    function generate_op_token(op) {
        return {
            type: "OPERATOR" /* OPERATOR */,
            value: op
        };
    }
    function read_end(endReg) {
        let str = "";
        let is_ended = ch => endReg.test(ch);
        while (!input.eof()) {
            let ch = input.peek();
            if (is_ended(ch)) {
                break;
            }
            else {
                input.next();
                str += ch;
            }
        }
        return str.trim();
    }
    function read_string() {
        /**
         * '#' end eg:
         * .icon-#{$size} {}
         */
        let str = read_end(/[,;{}():#\s]/);
        return {
            type: "TEXT" /* TEXT */,
            value: str
        };
    }
    function maybe_comment(ch) {
        if (is_comment_char(input.peek())) {
            skip_comment();
            return read_next();
        }
        else {
            return generate_op_token(ch);
        }
    }
    function maybe_placeholder(ch) {
        if (is_base_char(input.peek())) {
            return {
                type: "PLACEHOLDER" /* PLACEHOLDER */,
                value: ch + read_while(is_base_char)
            };
        }
        else {
            return generate_op_token(ch);
        }
    }
    function maybe_calculate_op_token(chStr) {
        if (input.peek() === ' ') {
            return generate_op_token(chStr);
        }
        else {
            return {
                type: "TEXT" /* TEXT */,
                value: chStr + read_string().value
            };
        }
    }
    function maybe_comparison_op_token(chStr) {
        if (is_comparison_op_tokens(chStr)) {
            return generate_op_token(chStr);
        }
        else {
            return {
                type: "TEXT" /* TEXT */,
                value: chStr + read_string().value
            };
        }
    }
    function read_punc(ch) {
        return {
            type: "PUNC" /* PUNC */,
            value: ch
        };
    }
    function maybe_punc(ch) {
        if (input.peek() === '{') {
            return read_punc(ch);
        }
        else {
            return {
                type: "TEXT" /* TEXT */,
                value: ch + read_string().value
            };
        }
    }
    function read_next() {
        read_while(is_whitespace);
        if (input.eof())
            return null;
        let ch = input.peek();
        /**
         * comment //  contains op /
         * Todos:还未支持多行注释如：\/**\/
        */
        if (is_comment_char(ch))
            return maybe_comment(input.next());
        if (is_placeholder_start(ch))
            return maybe_placeholder(input.next()); // @extend %message-shared;  contains op '%'
        if (is_assign_char(ch))
            return read_assign_char();
        if (is_id_start(ch))
            return read_ident();
        if (is_punc(ch)) {
            if (ch === '#') {
                return maybe_punc(input.next());
            }
            return read_punc(input.next());
        }
        if (is_keyword_start(ch))
            return read_keyword(); //eg: @extend .message-shared;
        if (is_calculate_op_char(ch))
            return maybe_calculate_op_token(read_while(is_calculate_op_char));
        if (is_comparison_op_char(ch))
            return maybe_comparison_op_token(read_while(is_comparison_op_char));
        if (is_base_char(ch))
            return read_string();
        input.croak("Can't handle character: " + ch);
    }
    /**
     * reset coordination after peek value
     */
    function peek_next() {
        let coordination = input.getCoordination();
        let tok = read_next();
        input.setCoordination(coordination);
        return tok;
    }
    function peek() {
        return peek_next();
    }
    function next() {
        return read_next();
    }
    function eliminateWhitespace() {
        read_while(is_whitespace);
    }
    function eof() {
        return peek() == null;
    }
}

function defaultOnError(error) {
    throw error;
}
function createCompilerError(code, loc, messages, additionalMessage) {
    const msg = (messages || errorMessages)[code] + (additionalMessage || ``);
    const error = new SyntaxError(String(msg));
    error.code = code;
    error.loc = loc;
    return error;
}
const errorMessages = {
    // parse errors
    [0 /* UNKNONWN_TOKEN_TYPE */]: 'Unknown token type.',
    [1 /* INVALID_LOC_POSITION */]: 'Incorrect sourceLocation. start loc should be smaller than end loc',
    [2 /* EXPECT_TEXT_NODE_AFTER_OPERATOR_NODE */]: 'Expect text node after operator node',
};

/**
 *
 * @param {lex processed stream method} input
 */
function parse(input, options) {
    const { filename, source } = options;
    function emitError(code, loc = {
        start: input.getCoordination(),
        end: input.getCoordination(),
        filename
    }) {
        defaultOnError(createCompilerError(code, loc));
    }
    /**
     * end with ';' , eg:
     *
     * $boder : 1px solid red;
     *
     * end with ',' | ')' , eg:
     *
     * @mixin test($param1:1,$param2:2){} // assign expression in @mixin or $include
      */
    function default_right_end_condition() {
        return is_punc(';');
    }
    function set_call_params_args_assign_right_end_condition() {
        assign_right_end_condition = () => is_punc(',') || is_punc(')');
    }
    function reset_assign_right_end_condition() {
        assign_right_end_condition = default_right_end_condition;
    }
    function is_punc(ch) {
        let tok = input.peek();
        return tok && tok.type == "PUNC" /* PUNC */ && (!ch || tok.value == ch) && tok;
    }
    function is_kw(kw) {
        let tok = input.peek();
        return tok && tok.type == "KEYWORD" /* KEYWORD */ && (!kw || tok.value == kw) && tok;
    }
    function is_op(op) {
        let tok = input.peek();
        return tok && tok.type == "OPERATOR" /* OPERATOR */ && (!op || tok.value == op) && tok;
    }
    function is_assign() {
        let tok = input.peek();
        return tok && tok.type === "ASSIGN" /* ASSIGN */;
    }
    function skip_punc(ch) {
        if (is_punc(ch))
            input.next();
        // Todos: no mandatory skip for now
        // else {
        //     input.croak("Expecting punctuation: \"" + ch + "\"");
        // }
    }
    // todos: replace typescript any
    function delimited(start, stop, separator, parser) {
        let a = [], first = true;
        skip_punc(start);
        while (!input.eof()) {
            if (debug())
                break;
            if (is_punc(stop))
                break;
            if (first)
                first = false;
            else
                skip_punc(separator);
            if (is_punc(stop))
                break;
            a.push(parser());
        }
        skip_punc(stop);
        return a;
    }
    let assign_right_end_condition = default_right_end_condition;
    // type parseFunctionType = <T extends Node>(...args:any[]) => T
    function injectLoc(parseFn) {
        return function injectLocInner(...args) {
            input.eliminateWhitespace();
            let start = input.getCoordination(), ast = parseFn.apply(null, args), end = input.getCoordination();
            /**
             *  to resolve stacked injectLoc executed which will lead to  start equal end position
             * */
            // if(ast.loc && ast.loc!==locStub && ast.loc.end.offset>0){
            //     return ast;
            // }
            /**
             * patch ASSIGN , make left node start as start position
             * patch BINARY node position, make binary left most node start as binary node start position
             * */
            if (ast.type === "BINARY" /* BINARY */ || ast.type === "ASSIGN" /* ASSIGN */) {
                let left = args[0];
                while (left.type === "BINARY" /* BINARY */) {
                    left = left.left;
                }
                start = left.loc.start;
            }
            let loc = {
                start,
                end,
                filename
            };
            if (start.offset >= end.offset) {
                emitError(1 /* INVALID_LOC_POSITION */);
            }
            return {
                ...ast,
                loc,
            };
        };
    }
    const parse_error = injectLoc(function parse_error() {
        function parse_list(endCheck = () => !default_right_end_condition()) {
            let list = [];
            while (endCheck()) {
                list.push(dispatchParser());
            }
            return {
                type: "LIST" /* LIST */,
                value: list
            };
        }
        return {
            type: "ERROR" /* ERROR */,
            value: parse_list()
        };
    });
    const consumeNextTokenWithLoc = injectLoc(() => input.next());
    function dispatchParser() {
        if (is_kw('@extend')) {
            input.next();
            return parse_extend();
        }
        if (is_kw('@mixin')) {
            input.next();
            return parse_mixin();
        }
        if (is_kw('@include')) {
            input.next();
            return parse_include();
        }
        if (is_kw('@import')) {
            input.next();
            return parse_import();
        }
        if (is_kw('@if')) {
            input.next();
            return parse_if();
        }
        if (is_kw('@each')) {
            input.next();
            return parse_each();
        }
        if (is_kw('@error')) {
            input.next();
            return parse_error();
        }
        let tok = input.peek();
        if (tok.type === "VARIABLE" /* VARIABLE */ || tok.type === "PLACEHOLDER" /* PLACEHOLDER */) {
            return consumeNextTokenWithLoc();
        }
        if (tok.type === "PUNC" /* PUNC */) {
            if (tok.value === "#") {
                return maybe_key_var_wrapper();
            }
            return consumeNextTokenWithLoc();
        }
        if (tok.type === "TEXT" /* TEXT */) {
            return parse_consecutive_str();
        }
        emitError(0 /* UNKNONWN_TOKEN_TYPE */);
    }
    function parseAssign(left) {
        input.next();
        function parse_assign_right() {
            let right = [];
            while (!assign_right_end_condition()) {
                if (debug())
                    break;
                let result = maybe_binary(dispatchParser(), 0);
                right.push(result);
            }
            return right;
        }
        let right = createListNode(parse_assign_right());
        return createAssignStatement(left, right);
    }
    function maybe_binary(left, left_prec) {
        let tok = is_op();
        if (tok) {
            if (PRECEDENCE[tok.value] > left_prec) {
                tok = consumeNextTokenWithLoc(); //skip op , add loc
                let nextNode = consumeNextTokenWithLoc();
                if (nextNode.type !== "TEXT" /* TEXT */) {
                    emitError(2 /* EXPECT_TEXT_NODE_AFTER_OPERATOR_NODE */, nextNode.loc);
                }
                return maybe_binary({
                    type: "BINARY" /* BINARY */,
                    operator: tok,
                    left: left,
                    right: maybe_binary(nextNode, PRECEDENCE[tok.value]),
                    loc: {
                        start: left.loc.start,
                        end: nextNode.loc.end,
                        filename
                    }
                }, left_prec);
            }
        }
        return left;
    }
    function parse_child(selector) {
        let children = delimited("{", "}", ";", parse_statement);
        return {
            type: "CHILD" /* CHILD */,
            selector,
            children,
            loc: locStub
        };
    }
    function maybe_assign(exp) {
        let expr = exp();
        if (is_assign()) {
            return parseAssign(expr);
        }
        /**
         * handle selector may contain key_var, which should be resolved to list ast
         * .icon-#{$size} {}
        */
        if (is_punc('#')) {
            return maybe_assign(() => {
                return createListNode(expr.type === "LIST" /* LIST */ ?
                    expr.value.concat(dispatchParser())
                    : [expr].concat(dispatchParser()));
            });
        }
        if (is_punc('{')) {
            return parse_child(expr); //passin selector
        }
        return expr;
    }
    function parse_extend() {
        return {
            type: "EXTEND" /* EXTEND */,
            param: consumeNextTokenWithLoc(),
            loc: locStub
        };
    }
    function parse_block_statement() {
        let children = delimited("{", "}", ";", parse_statement);
        return {
            type: "BODY" /* BODY */,
            children,
            loc: locStub
        };
    }
    /**
     * Todos: add @content kw and include body
     */
    function parse_mixin() {
        let id = {
            type: "IDENTIFIER" /* IDENTIFIER */,
            value: input.next().value,
            loc: locStub
        }, params = [];
        if (!is_punc('{')) {
            /**
             * Support default params, which contains assign ':' symbol; which will be processed in parseAssign
             * @mixin replace-text($image,$x:default1, $y:default2) {
              */
            set_call_params_args_assign_right_end_condition();
            params = delimited('(', ')', ',', parse_statement);
        }
        reset_assign_right_end_condition();
        return {
            type: "MIXIN" /* MIXIN */,
            id,
            params,
            body: parse_block_statement(),
            loc: locStub
        };
    }
    function parse_include() {
        let id = {
            type: "IDENTIFIER" /* IDENTIFIER */,
            value: input.next().value,
            loc: locStub
        }, args = []; // @include mixin1;
        if (!is_punc(';')) {
            set_call_params_args_assign_right_end_condition();
            args = delimited('(', ')', ',', parse_statement); // extend like expr or call expr
        }
        reset_assign_right_end_condition();
        return {
            type: "INCLUDE" /* INCLUDE */,
            id,
            args,
            loc: locStub
        };
    }
    function parse_import() {
        function processFilenameExp(exp) {
            exp.value = exp.value.match(/^['"](.+)['"]$/)[1];
            return exp;
        }
        let delimitor = input.peek(), params = [];
        if (!delimitor.value.startsWith("'") && !delimitor.value.startsWith('"')) {
            input.croak(`@import expected with ' or " but encountered ${delimitor}`);
        }
        while (!is_punc(';')) { // @import 'foundation/code', 'foundation/lists';
            params.push(processFilenameExp(consumeNextTokenWithLoc()));
            if (is_punc(',')) {
                skip_punc(',');
            }
        }
        return {
            type: "IMPORT" /* IMPORT */,
            params,
            loc: locStub
        };
    }
    function parse_if() {
        let alternate = null, testExpression, blockStatement;
        testExpression = maybe_binary(dispatchParser(), 0);
        if (!is_punc('{')) {
            input.croak(`@if expect '{' but encountered '${consumeNextTokenWithLoc().value}'`);
        }
        blockStatement = parse_block_statement();
        if (is_kw('@else')) {
            input.next();
            let predictToken = input.peek();
            /**
             * check if it's a @else if  statement
              */
            if (predictToken.type === "TEXT" /* TEXT */ && predictToken.value === 'if') {
                input.next();
                alternate = parse_if();
            }
            else {
                alternate = parse_block_statement();
            }
        }
        return {
            type: "IFSTATEMENT" /* IFSTATEMENT */,
            test: testExpression,
            consequent: blockStatement,
            alternate: alternate,
            loc: locStub
        };
    }
    function parse_each() {
        let left = dispatchParser();
        /**
         * skip "in" expression
          */
        input.next();
        let right = dispatchParser();
        skip_punc('{');
        let blockStatement = parse_statement();
        skip_punc('}');
        return {
            type: "EACHSTATEMENT" /* EACHSTATEMENT */,
            left,
            right,
            body: blockStatement,
            loc: locStub
        };
    }
    function parse_consecutive_str() {
        function maybe_call(exp) {
            let expr = exp();
            if (is_punc('(')) {
                return {
                    ...expr,
                    type: "TEXT" /* TEXT */,
                    value: expr.value + '(' + delimited('(', ')', ',', parse_statement).map(expr => expr.value).join(',') + ')'
                };
            }
            return expr;
        }
        function read_while(predicate) {
            let tokens = [];
            while (!input.eof() && predicate(input.peek()))
                //to resolve test skew(20deg) rotate(20deg);
                tokens.push(maybe_call(consumeNextTokenWithLoc));
            return tokens;
        }
        let list = read_while(tok => tok.type === "TEXT" /* TEXT */), listNode = createListNode(list);
        // no nested ListNode for now, so we flatten and return TextNode
        return {
            type: "TEXT" /* TEXT */,
            value: fillWhitespace(list).map(tok => tok.value).join(''),
            loc: listNode.loc
        };
    }
    /**
     *
     * #{var}
     */
    function maybe_key_var_wrapper() {
        function parse_key_var_wrapper(varKeyStartLoc) {
            skip_punc('{');
            let node = consumeNextTokenWithLoc();
            if (node.type !== "VARIABLE" /* VARIABLE */) {
                input.croak(`${node} should be a variable which starts with '$'`);
            }
            skip_punc('}');
            return {
                ...node,
                type: "VAR_KEY" /* VAR_KEY */,
                loc: {
                    start: varKeyStartLoc,
                    end: node.loc.end,
                    filename
                }
            };
        }
        let token = consumeNextTokenWithLoc();
        if (is_punc('{')) {
            return parse_key_var_wrapper(token.loc.start);
        }
        /**
         * color: #1212; or #selector{}
          */
        let nextToken = consumeNextTokenWithLoc();
        if (nextToken.type !== "TEXT" /* TEXT */) {
            input.croak(`[maybe_key_var_wrapper]: expect str token but received ${nextToken.value}`);
        }
        return {
            type: "TEXT" /* TEXT */,
            value: '#' + nextToken.value,
            loc: {
                start: token.start,
                end: nextToken.end,
                filename
            }
        };
    }
    function parse_statement() {
        return maybe_assign(function () {
            return dispatchParser();
        });
    }
    function parse_prog() {
        let children = [];
        while (!input.eof()) {
            let result = parse_statement();
            children.push(result);
            if (!input.eof())
                skip_punc(";");
        }
        return {
            type: "RootNode" /* RootNode */,
            fileSourceMap: {
                [filename]: source
            },
            children,
            loc: locStub
        };
    }
    return parse_prog();
}

// import runtime_schema_check from './runtime_ast_check'
function baseParse(scss, options) {
    return parse(lex(input_stream(scss)), options);
}

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */

const intToCharMap = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".split("");

/**
 * Encode an integer in the range of 0 to 63 to a single base 64 digit.
 */
var encode = function(number) {
  if (0 <= number && number < intToCharMap.length) {
    return intToCharMap[number];
  }
  throw new TypeError("Must be between 0 and 63: " + number);
};

var base64 = {
	encode: encode
};

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 *
 * Based on the Base 64 VLQ implementation in Closure Compiler:
 * https://code.google.com/p/closure-compiler/source/browse/trunk/src/com/google/debugging/sourcemap/Base64VLQ.java
 *
 * Copyright 2011 The Closure Compiler Authors. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *  * Neither the name of Google Inc. nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



// A single base 64 digit can contain 6 bits of data. For the base 64 variable
// length quantities we use in the source map spec, the first bit is the sign,
// the next four bits are the actual value, and the 6th bit is the
// continuation bit. The continuation bit tells us whether there are more
// digits in this value following this digit.
//
//   Continuation
//   |    Sign
//   |    |
//   V    V
//   101011

const VLQ_BASE_SHIFT = 5;

// binary: 100000
const VLQ_BASE = 1 << VLQ_BASE_SHIFT;

// binary: 011111
const VLQ_BASE_MASK = VLQ_BASE - 1;

// binary: 100000
const VLQ_CONTINUATION_BIT = VLQ_BASE;

/**
 * Converts from a two-complement value to a value where the sign bit is
 * placed in the least significant bit.  For example, as decimals:
 *   1 becomes 2 (10 binary), -1 becomes 3 (11 binary)
 *   2 becomes 4 (100 binary), -2 becomes 5 (101 binary)
 */
function toVLQSigned(aValue) {
  return aValue < 0
    ? ((-aValue) << 1) + 1
    : (aValue << 1) + 0;
}

/**
 * Returns the base 64 VLQ encoded value.
 */
var encode$1 = function base64VLQ_encode(aValue) {
  let encoded = "";
  let digit;

  let vlq = toVLQSigned(aValue);

  do {
    digit = vlq & VLQ_BASE_MASK;
    vlq >>>= VLQ_BASE_SHIFT;
    if (vlq > 0) {
      // There are still more digits in this value, so we must make sure the
      // continuation bit is marked.
      digit |= VLQ_CONTINUATION_BIT;
    }
    encoded += base64.encode(digit);
  } while (vlq > 0);

  return encoded;
};

var base64Vlq = {
	encode: encode$1
};

function createCommonjsModule(fn, basedir, module) {
	return module = {
	  path: basedir,
	  exports: {},
	  require: function (path, base) {
      return commonjsRequire(path, (base === undefined || base === null) ? module.path : base);
    }
	}, fn(module, module.exports), module.exports;
}

function commonjsRequire () {
	throw new Error('Dynamic requires are not currently supported by @rollup/plugin-commonjs');
}

var util = createCommonjsModule(function (module, exports) {
/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */

/**
 * This is a helper function for getting values from parameter/options
 * objects.
 *
 * @param args The object we are extracting values from
 * @param name The name of the property we are getting.
 * @param defaultValue An optional value to return if the property is missing
 * from the object. If this is not specified and the property is missing, an
 * error will be thrown.
 */
function getArg(aArgs, aName, aDefaultValue) {
  if (aName in aArgs) {
    return aArgs[aName];
  } else if (arguments.length === 3) {
    return aDefaultValue;
  }
    throw new Error('"' + aName + '" is a required argument.');

}
exports.getArg = getArg;

const urlRegexp = /^(?:([\w+\-.]+):)?\/\/(?:(\w+:\w+)@)?([\w.-]*)(?::(\d+))?(.*)$/;
const dataUrlRegexp = /^data:.+\,.+$/;

function urlParse(aUrl) {
  const match = aUrl.match(urlRegexp);
  if (!match) {
    return null;
  }
  return {
    scheme: match[1],
    auth: match[2],
    host: match[3],
    port: match[4],
    path: match[5]
  };
}
exports.urlParse = urlParse;

function urlGenerate(aParsedUrl) {
  let url = "";
  if (aParsedUrl.scheme) {
    url += aParsedUrl.scheme + ":";
  }
  url += "//";
  if (aParsedUrl.auth) {
    url += aParsedUrl.auth + "@";
  }
  if (aParsedUrl.host) {
    url += aParsedUrl.host;
  }
  if (aParsedUrl.port) {
    url += ":" + aParsedUrl.port;
  }
  if (aParsedUrl.path) {
    url += aParsedUrl.path;
  }
  return url;
}
exports.urlGenerate = urlGenerate;

const MAX_CACHED_INPUTS = 32;

/**
 * Takes some function `f(input) -> result` and returns a memoized version of
 * `f`.
 *
 * We keep at most `MAX_CACHED_INPUTS` memoized results of `f` alive. The
 * memoization is a dumb-simple, linear least-recently-used cache.
 */
function lruMemoize(f) {
  const cache = [];

  return function(input) {
    for (let i = 0; i < cache.length; i++) {
      if (cache[i].input === input) {
        const temp = cache[0];
        cache[0] = cache[i];
        cache[i] = temp;
        return cache[0].result;
      }
    }

    const result = f(input);

    cache.unshift({
      input,
      result,
    });

    if (cache.length > MAX_CACHED_INPUTS) {
      cache.pop();
    }

    return result;
  };
}

/**
 * Normalizes a path, or the path portion of a URL:
 *
 * - Replaces consecutive slashes with one slash.
 * - Removes unnecessary '.' parts.
 * - Removes unnecessary '<dir>/..' parts.
 *
 * Based on code in the Node.js 'path' core module.
 *
 * @param aPath The path or url to normalize.
 */
const normalize = lruMemoize(function normalize(aPath) {
  let path = aPath;
  const url = urlParse(aPath);
  if (url) {
    if (!url.path) {
      return aPath;
    }
    path = url.path;
  }
  const isAbsolute = exports.isAbsolute(path);

  // Split the path into parts between `/` characters. This is much faster than
  // using `.split(/\/+/g)`.
  const parts = [];
  let start = 0;
  let i = 0;
  while (true) {
    start = i;
    i = path.indexOf("/", start);
    if (i === -1) {
      parts.push(path.slice(start));
      break;
    } else {
      parts.push(path.slice(start, i));
      while (i < path.length && path[i] === "/") {
        i++;
      }
    }
  }

  let up = 0;
  for (i = parts.length - 1; i >= 0; i--) {
    const part = parts[i];
    if (part === ".") {
      parts.splice(i, 1);
    } else if (part === "..") {
      up++;
    } else if (up > 0) {
      if (part === "") {
        // The first part is blank if the path is absolute. Trying to go
        // above the root is a no-op. Therefore we can remove all '..' parts
        // directly after the root.
        parts.splice(i + 1, up);
        up = 0;
      } else {
        parts.splice(i, 2);
        up--;
      }
    }
  }
  path = parts.join("/");

  if (path === "") {
    path = isAbsolute ? "/" : ".";
  }

  if (url) {
    url.path = path;
    return urlGenerate(url);
  }
  return path;
});
exports.normalize = normalize;

/**
 * Joins two paths/URLs.
 *
 * @param aRoot The root path or URL.
 * @param aPath The path or URL to be joined with the root.
 *
 * - If aPath is a URL or a data URI, aPath is returned, unless aPath is a
 *   scheme-relative URL: Then the scheme of aRoot, if any, is prepended
 *   first.
 * - Otherwise aPath is a path. If aRoot is a URL, then its path portion
 *   is updated with the result and aRoot is returned. Otherwise the result
 *   is returned.
 *   - If aPath is absolute, the result is aPath.
 *   - Otherwise the two paths are joined with a slash.
 * - Joining for example 'http://' and 'www.example.com' is also supported.
 */
function join(aRoot, aPath) {
  if (aRoot === "") {
    aRoot = ".";
  }
  if (aPath === "") {
    aPath = ".";
  }
  const aPathUrl = urlParse(aPath);
  const aRootUrl = urlParse(aRoot);
  if (aRootUrl) {
    aRoot = aRootUrl.path || "/";
  }

  // `join(foo, '//www.example.org')`
  if (aPathUrl && !aPathUrl.scheme) {
    if (aRootUrl) {
      aPathUrl.scheme = aRootUrl.scheme;
    }
    return urlGenerate(aPathUrl);
  }

  if (aPathUrl || aPath.match(dataUrlRegexp)) {
    return aPath;
  }

  // `join('http://', 'www.example.com')`
  if (aRootUrl && !aRootUrl.host && !aRootUrl.path) {
    aRootUrl.host = aPath;
    return urlGenerate(aRootUrl);
  }

  const joined = aPath.charAt(0) === "/"
    ? aPath
    : normalize(aRoot.replace(/\/+$/, "") + "/" + aPath);

  if (aRootUrl) {
    aRootUrl.path = joined;
    return urlGenerate(aRootUrl);
  }
  return joined;
}
exports.join = join;

exports.isAbsolute = function(aPath) {
  return aPath.charAt(0) === "/" || urlRegexp.test(aPath);
};

/**
 * Make a path relative to a URL or another path.
 *
 * @param aRoot The root path or URL.
 * @param aPath The path or URL to be made relative to aRoot.
 */
function relative(aRoot, aPath) {
  if (aRoot === "") {
    aRoot = ".";
  }

  aRoot = aRoot.replace(/\/$/, "");

  // It is possible for the path to be above the root. In this case, simply
  // checking whether the root is a prefix of the path won't work. Instead, we
  // need to remove components from the root one by one, until either we find
  // a prefix that fits, or we run out of components to remove.
  let level = 0;
  while (aPath.indexOf(aRoot + "/") !== 0) {
    const index = aRoot.lastIndexOf("/");
    if (index < 0) {
      return aPath;
    }

    // If the only part of the root that is left is the scheme (i.e. http://,
    // file:///, etc.), one or more slashes (/), or simply nothing at all, we
    // have exhausted all components, so the path is not relative to the root.
    aRoot = aRoot.slice(0, index);
    if (aRoot.match(/^([^\/]+:\/)?\/*$/)) {
      return aPath;
    }

    ++level;
  }

  // Make sure we add a "../" for each component we removed from the root.
  return Array(level + 1).join("../") + aPath.substr(aRoot.length + 1);
}
exports.relative = relative;

const supportsNullProto = (function() {
  const obj = Object.create(null);
  return !("__proto__" in obj);
}());

function identity(s) {
  return s;
}

/**
 * Because behavior goes wacky when you set `__proto__` on objects, we
 * have to prefix all the strings in our set with an arbitrary character.
 *
 * See https://github.com/mozilla/source-map/pull/31 and
 * https://github.com/mozilla/source-map/issues/30
 *
 * @param String aStr
 */
function toSetString(aStr) {
  if (isProtoString(aStr)) {
    return "$" + aStr;
  }

  return aStr;
}
exports.toSetString = supportsNullProto ? identity : toSetString;

function fromSetString(aStr) {
  if (isProtoString(aStr)) {
    return aStr.slice(1);
  }

  return aStr;
}
exports.fromSetString = supportsNullProto ? identity : fromSetString;

function isProtoString(s) {
  if (!s) {
    return false;
  }

  const length = s.length;

  if (length < 9 /* "__proto__".length */) {
    return false;
  }

  /* eslint-disable no-multi-spaces */
  if (s.charCodeAt(length - 1) !== 95  /* '_' */ ||
      s.charCodeAt(length - 2) !== 95  /* '_' */ ||
      s.charCodeAt(length - 3) !== 111 /* 'o' */ ||
      s.charCodeAt(length - 4) !== 116 /* 't' */ ||
      s.charCodeAt(length - 5) !== 111 /* 'o' */ ||
      s.charCodeAt(length - 6) !== 114 /* 'r' */ ||
      s.charCodeAt(length - 7) !== 112 /* 'p' */ ||
      s.charCodeAt(length - 8) !== 95  /* '_' */ ||
      s.charCodeAt(length - 9) !== 95  /* '_' */) {
    return false;
  }
  /* eslint-enable no-multi-spaces */

  for (let i = length - 10; i >= 0; i--) {
    if (s.charCodeAt(i) !== 36 /* '$' */) {
      return false;
    }
  }

  return true;
}

/**
 * Comparator between two mappings where the original positions are compared.
 *
 * Optionally pass in `true` as `onlyCompareGenerated` to consider two
 * mappings with the same original source/line/column, but different generated
 * line and column the same. Useful when searching for a mapping with a
 * stubbed out mapping.
 */
function compareByOriginalPositions(mappingA, mappingB, onlyCompareOriginal) {
  let cmp = strcmp(mappingA.source, mappingB.source);
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalLine - mappingB.originalLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalColumn - mappingB.originalColumn;
  if (cmp !== 0 || onlyCompareOriginal) {
    return cmp;
  }

  cmp = mappingA.generatedColumn - mappingB.generatedColumn;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.generatedLine - mappingB.generatedLine;
  if (cmp !== 0) {
    return cmp;
  }

  return strcmp(mappingA.name, mappingB.name);
}
exports.compareByOriginalPositions = compareByOriginalPositions;

/**
 * Comparator between two mappings with deflated source and name indices where
 * the generated positions are compared.
 *
 * Optionally pass in `true` as `onlyCompareGenerated` to consider two
 * mappings with the same generated line and column, but different
 * source/name/original line and column the same. Useful when searching for a
 * mapping with a stubbed out mapping.
 */
function compareByGeneratedPositionsDeflated(mappingA, mappingB, onlyCompareGenerated) {
  let cmp = mappingA.generatedLine - mappingB.generatedLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.generatedColumn - mappingB.generatedColumn;
  if (cmp !== 0 || onlyCompareGenerated) {
    return cmp;
  }

  cmp = strcmp(mappingA.source, mappingB.source);
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalLine - mappingB.originalLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalColumn - mappingB.originalColumn;
  if (cmp !== 0) {
    return cmp;
  }

  return strcmp(mappingA.name, mappingB.name);
}
exports.compareByGeneratedPositionsDeflated = compareByGeneratedPositionsDeflated;

function strcmp(aStr1, aStr2) {
  if (aStr1 === aStr2) {
    return 0;
  }

  if (aStr1 === null) {
    return 1; // aStr2 !== null
  }

  if (aStr2 === null) {
    return -1; // aStr1 !== null
  }

  if (aStr1 > aStr2) {
    return 1;
  }

  return -1;
}

/**
 * Comparator between two mappings with inflated source and name strings where
 * the generated positions are compared.
 */
function compareByGeneratedPositionsInflated(mappingA, mappingB) {
  let cmp = mappingA.generatedLine - mappingB.generatedLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.generatedColumn - mappingB.generatedColumn;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = strcmp(mappingA.source, mappingB.source);
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalLine - mappingB.originalLine;
  if (cmp !== 0) {
    return cmp;
  }

  cmp = mappingA.originalColumn - mappingB.originalColumn;
  if (cmp !== 0) {
    return cmp;
  }

  return strcmp(mappingA.name, mappingB.name);
}
exports.compareByGeneratedPositionsInflated = compareByGeneratedPositionsInflated;

/**
 * Strip any JSON XSSI avoidance prefix from the string (as documented
 * in the source maps specification), and then parse the string as
 * JSON.
 */
function parseSourceMapInput(str) {
  return JSON.parse(str.replace(/^\)]}'[^\n]*\n/, ""));
}
exports.parseSourceMapInput = parseSourceMapInput;

/**
 * Compute the URL of a source given the the source root, the source's
 * URL, and the source map's URL.
 */
function computeSourceURL(sourceRoot, sourceURL, sourceMapURL) {
  sourceURL = sourceURL || "";

  if (sourceRoot) {
    // This follows what Chrome does.
    if (sourceRoot[sourceRoot.length - 1] !== "/" && sourceURL[0] !== "/") {
      sourceRoot += "/";
    }
    // The spec says:
    //   Line 4: An optional source root, useful for relocating source
    //   files on a server or removing repeated values in the
    //   “sources” entry.  This value is prepended to the individual
    //   entries in the “source” field.
    sourceURL = sourceRoot + sourceURL;
  }

  // Historically, SourceMapConsumer did not take the sourceMapURL as
  // a parameter.  This mode is still somewhat supported, which is why
  // this code block is conditional.  However, it's preferable to pass
  // the source map URL to SourceMapConsumer, so that this function
  // can implement the source URL resolution algorithm as outlined in
  // the spec.  This block is basically the equivalent of:
  //    new URL(sourceURL, sourceMapURL).toString()
  // ... except it avoids using URL, which wasn't available in the
  // older releases of node still supported by this library.
  //
  // The spec says:
  //   If the sources are not absolute URLs after prepending of the
  //   “sourceRoot”, the sources are resolved relative to the
  //   SourceMap (like resolving script src in a html document).
  if (sourceMapURL) {
    const parsed = urlParse(sourceMapURL);
    if (!parsed) {
      throw new Error("sourceMapURL could not be parsed");
    }
    if (parsed.path) {
      // Strip the last path component, but keep the "/".
      const index = parsed.path.lastIndexOf("/");
      if (index >= 0) {
        parsed.path = parsed.path.substring(0, index + 1);
      }
    }
    sourceURL = join(urlGenerate(parsed), sourceURL);
  }

  return normalize(sourceURL);
}
exports.computeSourceURL = computeSourceURL;
});

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */

/**
 * A data structure which is a combination of an array and a set. Adding a new
 * member is O(1), testing for membership is O(1), and finding the index of an
 * element is O(1). Removing elements from the set is not supported. Only
 * strings are supported for membership.
 */
class ArraySet {
  constructor() {
    this._array = [];
    this._set = new Map();
  }

  /**
   * Static method for creating ArraySet instances from an existing array.
   */
  static fromArray(aArray, aAllowDuplicates) {
    const set = new ArraySet();
    for (let i = 0, len = aArray.length; i < len; i++) {
      set.add(aArray[i], aAllowDuplicates);
    }
    return set;
  }

  /**
   * Return how many unique items are in this ArraySet. If duplicates have been
   * added, than those do not count towards the size.
   *
   * @returns Number
   */
  size() {
    return this._set.size;
  }

  /**
   * Add the given string to this set.
   *
   * @param String aStr
   */
  add(aStr, aAllowDuplicates) {
    const isDuplicate = this.has(aStr);
    const idx = this._array.length;
    if (!isDuplicate || aAllowDuplicates) {
      this._array.push(aStr);
    }
    if (!isDuplicate) {
      this._set.set(aStr, idx);
    }
  }

  /**
   * Is the given string a member of this set?
   *
   * @param String aStr
   */
  has(aStr) {
      return this._set.has(aStr);
  }

  /**
   * What is the index of the given string in the array?
   *
   * @param String aStr
   */
  indexOf(aStr) {
    const idx = this._set.get(aStr);
    if (idx >= 0) {
        return idx;
    }
    throw new Error('"' + aStr + '" is not in the set.');
  }

  /**
   * What is the element at the given index?
   *
   * @param Number aIdx
   */
  at(aIdx) {
    if (aIdx >= 0 && aIdx < this._array.length) {
      return this._array[aIdx];
    }
    throw new Error("No element indexed by " + aIdx);
  }

  /**
   * Returns the array representation of this set (which has the proper indices
   * indicated by indexOf). Note that this is a copy of the internal array used
   * for storing the members so that no one can mess with internal state.
   */
  toArray() {
    return this._array.slice();
  }
}
var ArraySet_1 = ArraySet;

var arraySet = {
	ArraySet: ArraySet_1
};

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2014 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */



/**
 * Determine whether mappingB is after mappingA with respect to generated
 * position.
 */
function generatedPositionAfter(mappingA, mappingB) {
  // Optimized for most common case
  const lineA = mappingA.generatedLine;
  const lineB = mappingB.generatedLine;
  const columnA = mappingA.generatedColumn;
  const columnB = mappingB.generatedColumn;
  return lineB > lineA || lineB == lineA && columnB >= columnA ||
         util.compareByGeneratedPositionsInflated(mappingA, mappingB) <= 0;
}

/**
 * A data structure to provide a sorted view of accumulated mappings in a
 * performance conscious manner. It trades a negligible overhead in general
 * case for a large speedup in case of mappings being added in order.
 */
class MappingList {
  constructor() {
    this._array = [];
    this._sorted = true;
    // Serves as infimum
    this._last = {generatedLine: -1, generatedColumn: 0};
  }

  /**
   * Iterate through internal items. This method takes the same arguments that
   * `Array.prototype.forEach` takes.
   *
   * NOTE: The order of the mappings is NOT guaranteed.
   */
  unsortedForEach(aCallback, aThisArg) {
    this._array.forEach(aCallback, aThisArg);
  }

  /**
   * Add the given source mapping.
   *
   * @param Object aMapping
   */
  add(aMapping) {
    if (generatedPositionAfter(this._last, aMapping)) {
      this._last = aMapping;
      this._array.push(aMapping);
    } else {
      this._sorted = false;
      this._array.push(aMapping);
    }
  }

  /**
   * Returns the flat, sorted array of mappings. The mappings are sorted by
   * generated position.
   *
   * WARNING: This method returns internal data without copying, for
   * performance. The return value must NOT be mutated, and should be treated as
   * an immutable borrow. If you want to take ownership, you must make your own
   * copy.
   */
  toArray() {
    if (!this._sorted) {
      this._array.sort(util.compareByGeneratedPositionsInflated);
      this._sorted = true;
    }
    return this._array;
  }
}

var MappingList_1 = MappingList;

var mappingList = {
	MappingList: MappingList_1
};

/* -*- Mode: js; js-indent-level: 2; -*- */
/*
 * Copyright 2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE or:
 * http://opensource.org/licenses/BSD-3-Clause
 */



const ArraySet$1 = arraySet.ArraySet;
const MappingList$1 = mappingList.MappingList;

/**
 * An instance of the SourceMapGenerator represents a source map which is
 * being built incrementally. You may pass an object with the following
 * properties:
 *
 *   - file: The filename of the generated source.
 *   - sourceRoot: A root for all relative URLs in this source map.
 */
class SourceMapGenerator {
  constructor(aArgs) {
    if (!aArgs) {
      aArgs = {};
    }
    this._file = util.getArg(aArgs, "file", null);
    this._sourceRoot = util.getArg(aArgs, "sourceRoot", null);
    this._skipValidation = util.getArg(aArgs, "skipValidation", false);
    this._sources = new ArraySet$1();
    this._names = new ArraySet$1();
    this._mappings = new MappingList$1();
    this._sourcesContents = null;
  }

  /**
   * Creates a new SourceMapGenerator based on a SourceMapConsumer
   *
   * @param aSourceMapConsumer The SourceMap.
   */
  static fromSourceMap(aSourceMapConsumer) {
    const sourceRoot = aSourceMapConsumer.sourceRoot;
    const generator = new SourceMapGenerator({
      file: aSourceMapConsumer.file,
      sourceRoot
    });
    aSourceMapConsumer.eachMapping(function(mapping) {
      const newMapping = {
        generated: {
          line: mapping.generatedLine,
          column: mapping.generatedColumn
        }
      };

      if (mapping.source != null) {
        newMapping.source = mapping.source;
        if (sourceRoot != null) {
          newMapping.source = util.relative(sourceRoot, newMapping.source);
        }

        newMapping.original = {
          line: mapping.originalLine,
          column: mapping.originalColumn
        };

        if (mapping.name != null) {
          newMapping.name = mapping.name;
        }
      }

      generator.addMapping(newMapping);
    });
    aSourceMapConsumer.sources.forEach(function(sourceFile) {
      let sourceRelative = sourceFile;
      if (sourceRoot !== null) {
        sourceRelative = util.relative(sourceRoot, sourceFile);
      }

      if (!generator._sources.has(sourceRelative)) {
        generator._sources.add(sourceRelative);
      }

      const content = aSourceMapConsumer.sourceContentFor(sourceFile);
      if (content != null) {
        generator.setSourceContent(sourceFile, content);
      }
    });
    return generator;
  }

  /**
   * Add a single mapping from original source line and column to the generated
   * source's line and column for this source map being created. The mapping
   * object should have the following properties:
   *
   *   - generated: An object with the generated line and column positions.
   *   - original: An object with the original line and column positions.
   *   - source: The original source file (relative to the sourceRoot).
   *   - name: An optional original token name for this mapping.
   */
  addMapping(aArgs) {
    const generated = util.getArg(aArgs, "generated");
    const original = util.getArg(aArgs, "original", null);
    let source = util.getArg(aArgs, "source", null);
    let name = util.getArg(aArgs, "name", null);

    if (!this._skipValidation) {
      this._validateMapping(generated, original, source, name);
    }

    if (source != null) {
      source = String(source);
      if (!this._sources.has(source)) {
        this._sources.add(source);
      }
    }

    if (name != null) {
      name = String(name);
      if (!this._names.has(name)) {
        this._names.add(name);
      }
    }

    this._mappings.add({
      generatedLine: generated.line,
      generatedColumn: generated.column,
      originalLine: original != null && original.line,
      originalColumn: original != null && original.column,
      source,
      name
    });
  }

  /**
   * Set the source content for a source file.
   */
  setSourceContent(aSourceFile, aSourceContent) {
    let source = aSourceFile;
    if (this._sourceRoot != null) {
      source = util.relative(this._sourceRoot, source);
    }

    if (aSourceContent != null) {
      // Add the source content to the _sourcesContents map.
      // Create a new _sourcesContents map if the property is null.
      if (!this._sourcesContents) {
        this._sourcesContents = Object.create(null);
      }
      this._sourcesContents[util.toSetString(source)] = aSourceContent;
    } else if (this._sourcesContents) {
      // Remove the source file from the _sourcesContents map.
      // If the _sourcesContents map is empty, set the property to null.
      delete this._sourcesContents[util.toSetString(source)];
      if (Object.keys(this._sourcesContents).length === 0) {
        this._sourcesContents = null;
      }
    }
  }

  /**
   * Applies the mappings of a sub-source-map for a specific source file to the
   * source map being generated. Each mapping to the supplied source file is
   * rewritten using the supplied source map. Note: The resolution for the
   * resulting mappings is the minimium of this map and the supplied map.
   *
   * @param aSourceMapConsumer The source map to be applied.
   * @param aSourceFile Optional. The filename of the source file.
   *        If omitted, SourceMapConsumer's file property will be used.
   * @param aSourceMapPath Optional. The dirname of the path to the source map
   *        to be applied. If relative, it is relative to the SourceMapConsumer.
   *        This parameter is needed when the two source maps aren't in the same
   *        directory, and the source map to be applied contains relative source
   *        paths. If so, those relative source paths need to be rewritten
   *        relative to the SourceMapGenerator.
   */
  applySourceMap(aSourceMapConsumer, aSourceFile, aSourceMapPath) {
    let sourceFile = aSourceFile;
    // If aSourceFile is omitted, we will use the file property of the SourceMap
    if (aSourceFile == null) {
      if (aSourceMapConsumer.file == null) {
        throw new Error(
          "SourceMapGenerator.prototype.applySourceMap requires either an explicit source file, " +
          'or the source map\'s "file" property. Both were omitted.'
        );
      }
      sourceFile = aSourceMapConsumer.file;
    }
    const sourceRoot = this._sourceRoot;
    // Make "sourceFile" relative if an absolute Url is passed.
    if (sourceRoot != null) {
      sourceFile = util.relative(sourceRoot, sourceFile);
    }
    // Applying the SourceMap can add and remove items from the sources and
    // the names array.
    const newSources = this._mappings.toArray().length > 0
      ? new ArraySet$1()
      : this._sources;
    const newNames = new ArraySet$1();

    // Find mappings for the "sourceFile"
    this._mappings.unsortedForEach(function(mapping) {
      if (mapping.source === sourceFile && mapping.originalLine != null) {
        // Check if it can be mapped by the source map, then update the mapping.
        const original = aSourceMapConsumer.originalPositionFor({
          line: mapping.originalLine,
          column: mapping.originalColumn
        });
        if (original.source != null) {
          // Copy mapping
          mapping.source = original.source;
          if (aSourceMapPath != null) {
            mapping.source = util.join(aSourceMapPath, mapping.source);
          }
          if (sourceRoot != null) {
            mapping.source = util.relative(sourceRoot, mapping.source);
          }
          mapping.originalLine = original.line;
          mapping.originalColumn = original.column;
          if (original.name != null) {
            mapping.name = original.name;
          }
        }
      }

      const source = mapping.source;
      if (source != null && !newSources.has(source)) {
        newSources.add(source);
      }

      const name = mapping.name;
      if (name != null && !newNames.has(name)) {
        newNames.add(name);
      }

    }, this);
    this._sources = newSources;
    this._names = newNames;

    // Copy sourcesContents of applied map.
    aSourceMapConsumer.sources.forEach(function(srcFile) {
      const content = aSourceMapConsumer.sourceContentFor(srcFile);
      if (content != null) {
        if (aSourceMapPath != null) {
          srcFile = util.join(aSourceMapPath, srcFile);
        }
        if (sourceRoot != null) {
          srcFile = util.relative(sourceRoot, srcFile);
        }
        this.setSourceContent(srcFile, content);
      }
    }, this);
  }

  /**
   * A mapping can have one of the three levels of data:
   *
   *   1. Just the generated position.
   *   2. The Generated position, original position, and original source.
   *   3. Generated and original position, original source, as well as a name
   *      token.
   *
   * To maintain consistency, we validate that any new mapping being added falls
   * in to one of these categories.
   */
  _validateMapping(aGenerated, aOriginal, aSource, aName) {
    // When aOriginal is truthy but has empty values for .line and .column,
    // it is most likely a programmer error. In this case we throw a very
    // specific error message to try to guide them the right way.
    // For example: https://github.com/Polymer/polymer-bundler/pull/519
    if (aOriginal && typeof aOriginal.line !== "number" && typeof aOriginal.column !== "number") {
        throw new Error(
            "original.line and original.column are not numbers -- you probably meant to omit " +
            "the original mapping entirely and only map the generated position. If so, pass " +
            "null for the original mapping instead of an object with empty or null values."
        );
    }

    if (aGenerated && "line" in aGenerated && "column" in aGenerated
        && aGenerated.line > 0 && aGenerated.column >= 0
        && !aOriginal && !aSource && !aName) ; else if (aGenerated && "line" in aGenerated && "column" in aGenerated
             && aOriginal && "line" in aOriginal && "column" in aOriginal
             && aGenerated.line > 0 && aGenerated.column >= 0
             && aOriginal.line > 0 && aOriginal.column >= 0
             && aSource) ; else {
      throw new Error("Invalid mapping: " + JSON.stringify({
        generated: aGenerated,
        source: aSource,
        original: aOriginal,
        name: aName
      }));
    }
  }

  /**
   * Serialize the accumulated mappings in to the stream of base 64 VLQs
   * specified by the source map format.
   */
  _serializeMappings() {
    let previousGeneratedColumn = 0;
    let previousGeneratedLine = 1;
    let previousOriginalColumn = 0;
    let previousOriginalLine = 0;
    let previousName = 0;
    let previousSource = 0;
    let result = "";
    let next;
    let mapping;
    let nameIdx;
    let sourceIdx;

    const mappings = this._mappings.toArray();
    for (let i = 0, len = mappings.length; i < len; i++) {
      mapping = mappings[i];
      next = "";

      if (mapping.generatedLine !== previousGeneratedLine) {
        previousGeneratedColumn = 0;
        while (mapping.generatedLine !== previousGeneratedLine) {
          next += ";";
          previousGeneratedLine++;
        }
      } else if (i > 0) {
        if (!util.compareByGeneratedPositionsInflated(mapping, mappings[i - 1])) {
          continue;
        }
        next += ",";
      }

      next += base64Vlq.encode(mapping.generatedColumn
                                 - previousGeneratedColumn);
      previousGeneratedColumn = mapping.generatedColumn;

      if (mapping.source != null) {
        sourceIdx = this._sources.indexOf(mapping.source);
        next += base64Vlq.encode(sourceIdx - previousSource);
        previousSource = sourceIdx;

        // lines are stored 0-based in SourceMap spec version 3
        next += base64Vlq.encode(mapping.originalLine - 1
                                   - previousOriginalLine);
        previousOriginalLine = mapping.originalLine - 1;

        next += base64Vlq.encode(mapping.originalColumn
                                   - previousOriginalColumn);
        previousOriginalColumn = mapping.originalColumn;

        if (mapping.name != null) {
          nameIdx = this._names.indexOf(mapping.name);
          next += base64Vlq.encode(nameIdx - previousName);
          previousName = nameIdx;
        }
      }

      result += next;
    }

    return result;
  }

  _generateSourcesContent(aSources, aSourceRoot) {
    return aSources.map(function(source) {
      if (!this._sourcesContents) {
        return null;
      }
      if (aSourceRoot != null) {
        source = util.relative(aSourceRoot, source);
      }
      const key = util.toSetString(source);
      return Object.prototype.hasOwnProperty.call(this._sourcesContents, key)
        ? this._sourcesContents[key]
        : null;
    }, this);
  }

  /**
   * Externalize the source map.
   */
  toJSON() {
    const map = {
      version: this._version,
      sources: this._sources.toArray(),
      names: this._names.toArray(),
      mappings: this._serializeMappings()
    };
    if (this._file != null) {
      map.file = this._file;
    }
    if (this._sourceRoot != null) {
      map.sourceRoot = this._sourceRoot;
    }
    if (this._sourcesContents) {
      map.sourcesContent = this._generateSourcesContent(map.sources, map.sourceRoot);
    }

    return map;
  }

  /**
   * Render the source map being generated to a string.
   */
  toString() {
    return JSON.stringify(this.toJSON());
  }
}

SourceMapGenerator.prototype._version = 3;
var SourceMapGenerator_1 = SourceMapGenerator;

var sourceMapGenerator = {
	SourceMapGenerator: SourceMapGenerator_1
};

/*
 * Copyright 2009-2011 Mozilla Foundation and contributors
 * Licensed under the New BSD license. See LICENSE.txt or:
 * http://opensource.org/licenses/BSD-3-Clause
 */
var SourceMapGenerator$1 = sourceMapGenerator.SourceMapGenerator;

function createCodegenContext(ast, { sourceMap = false }) {
    const context = {
        code: '',
        sourceMap,
        column: 1,
        line: 1,
        offset: 0,
        indentLevel: 0,
        // source: ast.source,
        push(code, node) {
            context.code += code;
            if ( context.map) {
                if (node) {
                    addMapping(node.loc.start, node.loc.filename);
                }
                advancePositionWithMutation(context, code);
                // end info is not needed for now ,which could compress source-map size
                // if(node) {
                //     addMapping(node.loc.end)
                // }
            }
        },
        indent() {
            newline(++context.indentLevel);
        },
        deindent(withoutNewLine = false) {
            if (withoutNewLine) {
                --context.indentLevel;
            }
            else {
                newline(--context.indentLevel);
            }
        },
        newline() {
            newline(context.indentLevel);
        }
    };
    function newline(n) {
        context.push('\n' + `  `.repeat(n));
    }
    function addMapping(loc, filename, name) {
        context.map.addMapping({
            name,
            source: filename,
            original: {
                line: loc.line,
                column: loc.column - 1
            },
            generated: {
                line: context.line,
                column: context.column - 1
            }
        });
    }
    if ( sourceMap) {
        context.map = new SourceMapGenerator$1();
        Object.keys(ast.fileSourceMap).forEach(filename => context.map.setSourceContent(filename, ast.fileSourceMap[filename]));
    }
    return context;
}
function generate(ast, options = {}) {
    const context = createCodegenContext(ast, options);
    ast.children.forEach((node) => genNode(node, context));
    return {
        ast,
        code: context.code,
        // SourceMapGenerator does have toJSON() method but it's not in the types
        map: context.map ? context.map.toJSON() : undefined
    };
}
function genNode(node, context) {
    if (!node) {
        console.error('node', node, 'context', context);
        return;
    }
    switch (node.type) {
        case "TEXT" /* TEXT */:
            genText(node, context);
            break;
        case "ASSIGN" /* ASSIGN */:
            genAssign(node, context);
            break;
        case "CHILD" /* CHILD */:
            genChild(node, context);
            break;
        case "EMPTY" /* EMPTY */:
            break;
        default:
            throw new Error("Don't know how to genNode for " + JSON.stringify(node));
    }
}
function genText(node, context) {
    context.push(node.value, node);
}
function genAssign(node, context) {
    const { push } = context;
    genNode(node.left, context);
    push(':');
    genNode(node.right, context);
    push(';');
}
function genChild(node, context) {
    const { push, deindent, indent, newline } = context;
    genNode(node.selector, context);
    push('{');
    indent();
    node.children.forEach((node, index) => {
        if (index && node.type !== "EMPTY" /* EMPTY */) {
            newline();
        }
        genNode(node, context);
    });
    deindent();
    push('}');
    newline();
}

function tranform_nest(ast) {
    function flatten_child(child, arr = []) {
        function flatten(child, parentSelector = '') {
            /**
             * https://sass-lang.com/documentation/style-rules/parent-selector
             * support parent selector with loosening restriction
             */
            let containParentSelector = false;
            let selector = child.selector;
            if (selector.value.indexOf('&') >= 0) {
                containParentSelector = true;
                selector.value = selector.value.replace(/&/, parentSelector);
            }
            if (!containParentSelector) {
                selector.value = (parentSelector + ' ' + selector.value).trim();
            }
            /**
             * 清理空的选择器
              */
            child.selector.value && arr.push(child);
            /**
             *
              */
            child.children.forEach((exp, index) => {
                if (exp.type === "CHILD" /* CHILD */) {
                    child.children.splice(index, 1, createEmptyNode());
                    flatten(exp, selector.value);
                }
            });
            return arr;
        }
        return flatten(child);
    }
    function toplevel(ast) {
        let children = [];
        ast.children.forEach(exp => {
            if (exp.type === "CHILD" /* CHILD */) {
                children = children.concat(...flatten_child(exp));
            }
            else {
                children.push(exp);
            }
        });
        ast.children = children;
        return ast;
    }
    return toplevel(ast);
}

function transform_extend(ast) {
    /**
     * transform
     *
     * .basic{}
     * .primary{
     *     @extend .basic;
     * }
     *
     * to
     *
     * {'.basic':['primary']}
     */
    let extendSelectorPair = {};
    function is_placeholder(exp) {
        return exp.type === "PLACEHOLDER" /* PLACEHOLDER */;
    }
    function rm_empty_child(exp) {
        return exp.type === "CHILD" /* CHILD */ && (is_placeholder(exp.selector) || exp.children.length === 0) ? null : exp;
    }
    function collect_extend(exp) {
        function collect(child) {
            child.children = child.children.map(exp => {
                if (exp.type === "EXTEND" /* EXTEND */) {
                    extendSelectorPair[exp.param.value] = (extendSelectorPair[exp.param.value] || []).concat(child.selector.value);
                    return null;
                }
                return exp;
            }).filter(exp => exp !== null);
            return child;
        }
        return exp.type === "CHILD" /* CHILD */ ? collect(exp) : exp;
    }
    function transform_extend(exp) {
        function transform(child) {
            if (is_placeholder(child.selector)) {
                child.selector = {
                    type: "TEXT" /* TEXT */,
                    value: extendSelectorPair[child.selector.value].join(','),
                    loc: child.selector.loc
                };
            }
            else {
                child.selector = {
                    type: "TEXT" /* TEXT */,
                    value: extendSelectorPair[child.selector.value].concat(child.selector.value).join(','),
                    loc: child.selector.loc
                };
            }
            return child;
        }
        return exp.type === "CHILD" /* CHILD */ && extendSelectorPair[exp.selector.value] ? transform(exp) : exp;
    }
    function toplevel(ast) {
        ast.children = ast.children.map(exp => collect_extend(exp));
        ast.children = ast.children.map(exp => transform_extend(exp));
        ast.children = ast.children.map(exp => rm_empty_child(exp)).filter(exp => exp !== null);
        return ast;
    }
    return toplevel(ast);
}

/**
 *
 * ast => ast
 * replace @import to real ast
 *
 */
const EXTNAME_GLOBAL = '.scss';
function transform_module(rootNode, options) {
    const { sourceDir = './' } = options;
    let statementList = [];
    /**
     * Todos
     * only support @import in the head lines for now
     *
     *  depth/left first walk import module, push child module children before parent module child
     */
    function walkNode(rootNode) {
        function importModule(module, parent) {
            module.params.forEach((param) => {
                let filename = param.value, extname = path.extname(filename), basename = path.basename(filename), dirname = path.dirname(filename), filePath = path.join(sourceDir, dirname, '_' + basename + (extname ? '' : EXTNAME_GLOBAL)), source = fs.readFileSync(filePath, 'utf8'), fileSourceMap = {
                    filename,
                    source
                };
                let childRootNode = baseParse(source, fileSourceMap);
                walkNode(childRootNode);
                // collect source map file in module
                Object.assign(parent.fileSourceMap, childRootNode.fileSourceMap);
            });
        }
        rootNode.children.forEach((statement) => {
            if (statement.type === "IMPORT" /* IMPORT */) {
                importModule(statement, rootNode);
            }
            else {
                statementList.push(statement);
            }
        });
    }
    walkNode(rootNode);
    rootNode.children = statementList;
    return rootNode;
}

let transformMiddleware = [tranform_nest, transform_extend];
var transformChain = (ast) => transformMiddleware.reduce((ast, middleware) => {
    return middleware.call(null, ast);
}, ast);

function createTransformContext(root, { nodeTransforms = [], onError = defaultOnError, sourceDir = './' }) {
    const context = {
        onError,
        nodeTransforms,
        root,
        sourceDir,
        env: new Environment(null),
        parent: null,
        currentNode: root,
        childIndex: 0,
        onNodeRemoved: () => { },
        replaceNode(node) {
            if (!context.parent) {
                throw new Error(`Cannot replace root node.`);
            }
            context.parent.children[context.childIndex] = context.currentNode = node;
            return node;
        }
    };
    return context;
}
function transform(root, options) {
    const context = createTransformContext(root, options);
    {
        root = transform_module(root, options);
    }
    traverseNode(root, context);
    // transformChain will be slowly replaced by transform plugins if possible
    root = transformChain(root);
}
function traverseChildren(parent, context) {
    let i = 0;
    const nodeRemoved = () => {
        i--;
    };
    for (; i < parent.children.length; i++) {
        const child = parent.children[i];
        context.parent = parent;
        context.childIndex = i;
        context.onNodeRemoved = nodeRemoved;
        traverseNode(child, context);
    }
    parent.children = parent.children.filter((node) => node.type !== "EMPTY" /* EMPTY */);
}
/*
only traverse top level for now
this sass compiler combined evaluate and compiling to direct target code so it needs to construct executing env while compiling
so it will be difficult for pure transform plugins
different from compile to target code which is evaluatable
  */
function traverseNode(node, context) {
    context.currentNode = node;
    // apply transform plugins
    const { nodeTransforms } = context;
    if (node.type !== "RootNode" /* RootNode */) {
        for (let i = 0; i < nodeTransforms.length; i++) {
            nodeTransforms[i](node, context);
        }
    }
    if (node.type === "RootNode" /* RootNode */) {
        traverseChildren(node, context);
    }
}

function processExpression(node, context) {
    function dispatchExpression(node, context) {
        switch (node.type) {
            /**
             * Expression
             */
            case "TEXT" /* TEXT */: return transformText(node);
            case "PUNC" /* PUNC */: return transformPunc(node);
            case "OPERATOR" /* OPERATOR */: return transformOp(node);
            case "VARIABLE" /* VARIABLE */: return transformVar(node, context);
            case "VAR_KEY" /* VAR_KEY */: return transformVarKey(node, context);
            case "LIST" /* LIST */: return transformList(node, context);
            case "BINARY" /* BINARY */: return transformBinary(node, context);
            default:
                throw new Error("Don't know how to processExpression expression for " + JSON.stringify(node));
        }
    }
    /**
     * Solve situation, assign value with NodeTypes.PUNC, eg:
     * $font: Helvetica, sans-serif;
     */
    function transformPunc(node) {
        return {
            ...node,
            type: "TEXT" /* TEXT */
        };
    }
    /**
     * Solve situation, selector with operators, eg:
     * .a > b{}
     */
    function transformOp(node) {
        return transformPunc(node);
    }
    function transformText(node) {
        return node;
    }
    function transformVar(node, context) {
        return {
            ...node,
            type: "TEXT" /* TEXT */,
            value: context.env.get(node.value)
        };
    }
    function transformVarKey(node, context) {
        return transformVar(node, context);
    }
    function transformList(node, context) {
        return {
            type: "TEXT" /* TEXT */,
            loc: node.loc,
            value: fillWhitespace(node.value).map((item) => {
                return dispatchExpression(item, context).value;
            }).join('').trim()
        };
    }
    /**
     * Todos:add more unit type eg: pt px etc
     *
     */
    function transformBinary(node, context) {
        let hasPercent = false, unitExtracted = false, unit = "";
        function parseFloatFn(str) {
            if (isNaN(parseFloat(str))) {
                return str;
            }
            if (str === "100%") {
                hasPercent = true;
                return 1;
            }
            /**
             * make the first encountered unit as the final unit
              */
            if (!unitExtracted) {
                unitExtracted = true;
                let matched = str.match(/\d+([a-z]*)/);
                unit = matched && matched.length > 0 ? matched[1] : '';
            }
            return parseFloat(str);
        }
        function transformVal(str) {
            function transformPercent(str) {
                return parseFloat(str) * 100 + "%";
            }
            function addUnit(str) {
                return str + unit;
            }
            return hasPercent ? transformPercent(str) : addUnit(str);
        }
        function evaluateBinary(node) {
            const opAcMap = {
                // '=': (left, right) => left = right,
                '||': (left, right) => left || right,
                '&&': (left, right) => left && right,
                '==': (left, right) => left == right,
                '!=': (left, right) => left != right,
                '>=': (left, right) => left >= right,
                '<=': (left, right) => left <= right,
                '+': (left, right) => left + right,
                '-': (left, right) => left - right,
                '/': (left, right) => left / right,
                '*': (left, right) => left * right,
                '%': (left, right) => left % right,
            };
            if (node.type === "TEXT" /* TEXT */)
                return parseFloatFn(node.value);
            if (node.type === "VARIABLE" /* VARIABLE */)
                return evaluateBinary({
                    ...node,
                    type: "TEXT" /* TEXT */,
                    value: context.env.get(node.value),
                });
            if (node.type === "BINARY" /* BINARY */)
                return opAcMap[node.operator.value](evaluateBinary(node.left), evaluateBinary(node.right));
            throw new Error("Don't know how to evaluateBinary type: " + node.type);
        }
        let value = evaluateBinary(node);
        return {
            ...node,
            type: "TEXT" /* TEXT */,
            value: transformVal(value),
        };
    }
    return dispatchExpression(node, context);
}

const processAssign = (node, context) => {
    if (node.left.type === "VARIABLE" /* VARIABLE */) {
        context.env.def(node.left.value, processExpression(node.right, context).value);
        return createEmptyNode();
    }
    if (node.left.type === "VAR_KEY" /* VAR_KEY */) {
        node.left = processExpression(node.left, context);
    }
    node.right = processExpression(node.right, context);
    return node;
};

const transformStatement = (node, context) => {
    processStatement(node, context);
};
function processStatement(node, context) {
    function dispatchStatement(node, context) {
        // console.log(context.env.parent)
        switch (node.type) {
            /**
            * Statement
            */
            case "ASSIGN" /* ASSIGN */: return processAssign(node, context);
            case "MIXIN" /* MIXIN */: return transform_mixin(node, context);
            case "INCLUDE" /* INCLUDE */: return transform_include(node, context);
            case "CHILD" /* CHILD */:
            case "BODY" /* BODY */:
                return transform_child_or_body(node, context);
            case "EXTEND" /* EXTEND */: return node;
            case "IFSTATEMENT" /* IFSTATEMENT */: return transform_if(node, context);
            case "EACHSTATEMENT" /* EACHSTATEMENT */: return transform_each(node, context);
            case "ERROR" /* ERROR */:
                throw new Error(processExpression(node.value, context).value);
            default:
                throw new Error("Don't know how to compile statement for " + JSON.stringify(node));
        }
    }
    /**
     * any expressions separated with spaces or commas count as a list
     */
    function transform_each(node, context) {
        let restoredContext = deepClone(node), right = processExpression(restoredContext.right, context), scope = context.env.extend(), list = [];
        if (right.type === "TEXT" /* TEXT */) {
            list = right.value.split(/,|\s+/).filter(val => val !== '');
        }
        let children = list.map(val => {
            /**
             * restore context with iterate multiple times
             */
            restoredContext = deepClone(node);
            scope.def(restoredContext.left.value, val);
            return dispatchStatement(restoredContext.body, { ...context, env: scope });
        });
        /**
         * return a created an empty selector child whose children will be flattened in transform_nest
         */
        let child = {
            ...node,
            type: "CHILD" /* CHILD */,
            selector: addNodeEmptyLocation({
                type: "TEXT" /* TEXT */,
                value: ''
            }),
            children
        };
        return child;
    }
    /**
     * context will be restored with function
     */
    function transform_if(node, context) {
        function is_true_exp(expression, context) {
            let resultExp = expression;
            /**
             *  1. if $bool { }
             *  2. if true { }
            */
            if (resultExp.type === "VARIABLE" /* VARIABLE */ || resultExp.type === "BINARY" /* BINARY */) {
                resultExp = processExpression(expression, context);
            }
            if (resultExp.value === "0" || resultExp.value === "false") {
                return false;
            }
            return true;
        }
        if (is_true_exp(node.test, context)) {
            return dispatchStatement(node.consequent, context);
        }
        else if (node.alternate) {
            return dispatchStatement(node.alternate, context);
        }
        else {
            return createEmptyNode();
        }
    }
    /**
     * transform @mixin -> set to env -> delete @mixin ast
     */
    function transform_mixin(node, context) {
        function make_function() {
            /**
             * deep clone function statement to restore context when called multiple times
             */
            let restoredContext = deepClone(node), params = restoredContext.params, scope = context.env.extend();
            function handle_params_default_value(params) {
                return params.map((param) => {
                    let ret = param;
                    if (param.type === "ASSIGN" /* ASSIGN */) {
                        ret = {
                            type: "VARIABLE" /* VARIABLE */,
                            value: param.left.value
                        };
                        dispatchStatement(param, { ...context, env: scope });
                    }
                    return ret;
                });
            }
            params = handle_params_default_value(params);
            params.forEach((param, i) => {
                if (param.type === "VARIABLE" /* VARIABLE */ && arguments[i]) {
                    scope.def(param.value, arguments[i]);
                }
            });
            return dispatchStatement(restoredContext.body, { ...context, env: scope });
        }
        context.env.def(node.id.value, make_function);
        return createEmptyNode();
    }
    function transform_include(node, context) {
        let func = context.env.get(node.id.value);
        return func.apply(null, node.args.map(arg => {
            /**
             * @include avatar(100px, $circle: false);
             */
            if (arg.type === "ASSIGN" /* ASSIGN */) {
                return processExpression(arg.right, context).value;
            }
            else {
                return processExpression(arg, context).value;
            }
        }));
    }
    function transform_child_or_body(node, context) {
        function flatten_included_body(children) {
            let arr = [];
            children.forEach(child => {
                if (child.type === "BODY" /* BODY */) {
                    arr = arr.concat(flatten_included_body(child.children));
                }
                else {
                    arr.push(child);
                }
            });
            return arr;
        }
        let scope = context.env.extend();
        node.children = node.children.map(child => dispatchStatement(child, { ...context, env: scope })).filter(node => !isEmptyNode(node));
        if (node.selector && node.selector.type === "LIST" /* LIST */) {
            node.selector = processExpression(node.selector, { ...context, env: scope });
        }
        /**
         * resolve BlockStatement/body
         *
         * @include include_function_body
         * @if which contain BlockStatement
          */
        node.children = flatten_included_body(node.children);
        return node;
    }
    return context.replaceNode(dispatchStatement(node, context));
}

function baseCompile(scss, options = {
    filename: 'default.scss',
    source: ''
}) {
    let ast = baseParse(scss, options);
    transform(ast, {
        nodeTransforms: [transformStatement],
        ...options
    });
    return generate(ast, options);
}

const { promisify } = require$$0;

const optsArg = opts => {
  if (!opts)
    opts = { mode: 0o777, fs };
  else if (typeof opts === 'object')
    opts = { mode: 0o777, fs, ...opts };
  else if (typeof opts === 'number')
    opts = { mode: opts, fs };
  else if (typeof opts === 'string')
    opts = { mode: parseInt(opts, 8), fs };
  else
    throw new TypeError('invalid options argument')

  opts.mkdir = opts.mkdir || opts.fs.mkdir || fs.mkdir;
  opts.mkdirAsync = promisify(opts.mkdir);
  opts.stat = opts.stat || opts.fs.stat || fs.stat;
  opts.statAsync = promisify(opts.stat);
  opts.statSync = opts.statSync || opts.fs.statSync || fs.statSync;
  opts.mkdirSync = opts.mkdirSync || opts.fs.mkdirSync || fs.mkdirSync;
  return opts
};
var optsArg_1 = optsArg;

const platform = process.env.__TESTING_MKDIRP_PLATFORM__ || process.platform;
const { resolve, parse: parse$1 } = path;
const pathArg = path => {
  if (/\0/.test(path)) {
    // simulate same failure that node raises
    throw Object.assign(
      new TypeError('path must be a string without null bytes'),
      {
        path,
        code: 'ERR_INVALID_ARG_VALUE',
      }
    )
  }

  path = resolve(path);
  if (platform === 'win32') {
    const badWinChars = /[*|"<>?:]/;
    const {root} = parse$1(path);
    if (badWinChars.test(path.substr(root.length))) {
      throw Object.assign(new Error('Illegal characters in path.'), {
        path,
        code: 'EINVAL',
      })
    }
  }

  return path
};
var pathArg_1 = pathArg;

const {dirname} = path;

const findMade = (opts, parent, path = undefined) => {
  // we never want the 'made' return value to be a root directory
  if (path === parent)
    return Promise.resolve()

  return opts.statAsync(parent).then(
    st => st.isDirectory() ? path : undefined, // will fail later
    er => er.code === 'ENOENT'
      ? findMade(opts, dirname(parent), parent)
      : undefined
  )
};

const findMadeSync = (opts, parent, path = undefined) => {
  if (path === parent)
    return undefined

  try {
    return opts.statSync(parent).isDirectory() ? path : undefined
  } catch (er) {
    return er.code === 'ENOENT'
      ? findMadeSync(opts, dirname(parent), parent)
      : undefined
  }
};

var findMade_1 = {findMade, findMadeSync};

const {dirname: dirname$1} = path;

const mkdirpManual = (path, opts, made) => {
  opts.recursive = false;
  const parent = dirname$1(path);
  if (parent === path) {
    return opts.mkdirAsync(path, opts).catch(er => {
      // swallowed by recursive implementation on posix systems
      // any other error is a failure
      if (er.code !== 'EISDIR')
        throw er
    })
  }

  return opts.mkdirAsync(path, opts).then(() => made || path, er => {
    if (er.code === 'ENOENT')
      return mkdirpManual(parent, opts)
        .then(made => mkdirpManual(path, opts, made))
    if (er.code !== 'EEXIST' && er.code !== 'EROFS')
      throw er
    return opts.statAsync(path).then(st => {
      if (st.isDirectory())
        return made
      else
        throw er
    }, () => { throw er })
  })
};

const mkdirpManualSync = (path, opts, made) => {
  const parent = dirname$1(path);
  opts.recursive = false;

  if (parent === path) {
    try {
      return opts.mkdirSync(path, opts)
    } catch (er) {
      // swallowed by recursive implementation on posix systems
      // any other error is a failure
      if (er.code !== 'EISDIR')
        throw er
      else
        return
    }
  }

  try {
    opts.mkdirSync(path, opts);
    return made || path
  } catch (er) {
    if (er.code === 'ENOENT')
      return mkdirpManualSync(path, opts, mkdirpManualSync(parent, opts, made))
    if (er.code !== 'EEXIST' && er.code !== 'EROFS')
      throw er
    try {
      if (!opts.statSync(path).isDirectory())
        throw er
    } catch (_) {
      throw er
    }
  }
};

var mkdirpManual_1 = {mkdirpManual, mkdirpManualSync};

const {dirname: dirname$2} = path;
const {findMade: findMade$1, findMadeSync: findMadeSync$1} = findMade_1;
const {mkdirpManual: mkdirpManual$1, mkdirpManualSync: mkdirpManualSync$1} = mkdirpManual_1;

const mkdirpNative = (path, opts) => {
  opts.recursive = true;
  const parent = dirname$2(path);
  if (parent === path)
    return opts.mkdirAsync(path, opts)

  return findMade$1(opts, path).then(made =>
    opts.mkdirAsync(path, opts).then(() => made)
    .catch(er => {
      if (er.code === 'ENOENT')
        return mkdirpManual$1(path, opts)
      else
        throw er
    }))
};

const mkdirpNativeSync = (path, opts) => {
  opts.recursive = true;
  const parent = dirname$2(path);
  if (parent === path)
    return opts.mkdirSync(path, opts)

  const made = findMadeSync$1(opts, path);
  try {
    opts.mkdirSync(path, opts);
    return made
  } catch (er) {
    if (er.code === 'ENOENT')
      return mkdirpManualSync$1(path, opts)
    else
      throw er
  }
};

var mkdirpNative_1 = {mkdirpNative, mkdirpNativeSync};

const version = process.env.__TESTING_MKDIRP_NODE_VERSION__ || process.version;
const versArr = version.replace(/^v/, '').split('.');
const hasNative = +versArr[0] > 10 || +versArr[0] === 10 && +versArr[1] >= 12;

const useNative = !hasNative ? () => false : opts => opts.mkdir === fs.mkdir;
const useNativeSync = !hasNative ? () => false : opts => opts.mkdirSync === fs.mkdirSync;

var useNative_1 = {useNative, useNativeSync};

const {mkdirpNative: mkdirpNative$1, mkdirpNativeSync: mkdirpNativeSync$1} = mkdirpNative_1;
const {mkdirpManual: mkdirpManual$2, mkdirpManualSync: mkdirpManualSync$2} = mkdirpManual_1;
const {useNative: useNative$1, useNativeSync: useNativeSync$1} = useNative_1;


const mkdirp = (path, opts) => {
  path = pathArg_1(path);
  opts = optsArg_1(opts);
  return useNative$1(opts)
    ? mkdirpNative$1(path, opts)
    : mkdirpManual$2(path, opts)
};

const mkdirpSync = (path, opts) => {
  path = pathArg_1(path);
  opts = optsArg_1(opts);
  return useNativeSync$1(opts)
    ? mkdirpNativeSync$1(path, opts)
    : mkdirpManualSync$2(path, opts)
};

mkdirp.sync = mkdirpSync;
mkdirp.native = (path, opts) => mkdirpNative$1(pathArg_1(path), optsArg_1(opts));
mkdirp.manual = (path, opts) => mkdirpManual$2(pathArg_1(path), optsArg_1(opts));
mkdirp.nativeSync = (path, opts) => mkdirpNativeSync$1(pathArg_1(path), optsArg_1(opts));
mkdirp.manualSync = (path, opts) => mkdirpManualSync$2(pathArg_1(path), optsArg_1(opts));

var mkdirp_1 = mkdirp;

function require_css(scssPath) {
    return {
        source: fs.readFileSync(scssPath, 'utf8'),
        filename: path.basename(scssPath)
    };
}
function run(sourceDir, outputDir = './', options = {
    genOtherInfo: false,
    sourceMap: false
}) {
    let sourceDirLength = path.basename(sourceDir).length;
    if (!fs.existsSync(outputDir)) {
        mkdirp_1.sync(outputDir);
    }
    function render(filePath) {
        if (path.extname(filePath) !== '.scss')
            return;
        let requireCss = require_css(filePath), basename = path.basename(filePath, '.scss'), sourceDirname = path.dirname(filePath), normalDistPath = path.join(outputDir, sourceDirname.substr(sourceDirLength)), astDistPath = path.join(outputDir, 'ast', sourceDirname.substr(sourceDirLength)), codeGenAstDistPath = path.join(outputDir, 'code-gen-ast', sourceDirname.substr(sourceDirLength)), sourceMapDistPath = path.join(outputDir, 'source-map', sourceDirname.substr(sourceDirLength)), cssDistPath = path.join(outputDir, 'css', sourceDirname.substr(sourceDirLength)), ast;
        if (!options.genOtherInfo) {
            if (!fs.existsSync(normalDistPath)) {
                mkdirp_1.sync(normalDistPath);
            }
            cssDistPath = normalDistPath;
        }
        else {
            if (!fs.existsSync(cssDistPath)) {
                mkdirp_1.sync(cssDistPath);
            }
            if (!fs.existsSync(codeGenAstDistPath)) {
                mkdirp_1.sync(codeGenAstDistPath);
            }
            if (!fs.existsSync(astDistPath)) {
                mkdirp_1.sync(astDistPath);
            }
            if (!fs.existsSync(sourceMapDistPath)) {
                mkdirp_1.sync(sourceMapDistPath);
            }
        }
        try {
            ast = baseParse(requireCss.source, requireCss);
        }
        catch (e) {
            console.error(`\nParser Error:\n filePath: ${filePath}\n`, e);
            return;
        }
        function write_parsed_ast(cb) {
            fs.writeFile(path.join(astDistPath, basename + '.json'), JSON.stringify(ast, null, 2), function (err) {
                if (err) {
                    console.error(err);
                    return console.error(`parse failed ${basename}`);
                }
                cb();
            });
        }
        function write_compiled() {
            let compiled;
            try {
                compiled = baseCompile(requireCss.source, {
                    ...requireCss,
                    sourceDir: sourceDirname,
                    sourceMap: options.sourceMap
                });
            }
            catch (e) {
                console.log('Error source code path: ', filePath);
                console.log(e);
                return;
            }
            const { ast, map, code } = compiled;
            function writeResultCode(cb) {
                fs.writeFile(path.join(cssDistPath, basename + '.css'), (code), function (err) {
                    if (err) {
                        return console.error(`write css failed ${basename}`);
                    }
                    cb();
                });
            }
            function writeCodegenAst(cb) {
                fs.writeFile(path.join(codeGenAstDistPath, basename + '.json'), JSON.stringify(ast, null, 2), function (err) {
                    if (err) {
                        return console.error(`write transformed ast failed ${basename},\nError:${err}`);
                    }
                    cb();
                });
            }
            function writeSourceMap() {
                fs.writeFile(path.join(sourceMapDistPath, basename + '.css.map'), JSON.stringify(map, null, 2), function (err) {
                    if (err) {
                        return console.error(`write source map failed ${basename}`);
                    }
                    console.log(`compile success ${basename}`);
                });
            }
            writeResultCode(() => {
                if (!options.genOtherInfo) {
                    return;
                }
                writeCodegenAst(() => {
                    writeSourceMap();
                });
            });
        }
        if (!options.genOtherInfo) {
            write_compiled();
        }
        else {
            write_parsed_ast(() => {
                write_compiled();
            });
        }
    }
    function renderDir(sourceDir) {
        fs.readdirSync(sourceDir).forEach((filename) => {
            //Todos Only analyze filename without '_'
            if (filename.startsWith('_')) {
                return;
            }
            // if (filename!== 'flow-control' &&  filename !== 'each' && filename !=='each.scss') return;
            // if (filename!== 'var-simple.scss') return;
            // if (filename!== 'operator.scss') return;
            let filePath = path.join(sourceDir, filename), stat = fs.lstatSync(filePath);
            if (stat.isFile()) {
                return render(filePath);
            }
            else if (stat.isDirectory()) {
                renderDir(filePath);
            }
            else {
                console.error('Unknown file type');
            }
        });
    }
    renderDir(sourceDir);
}

export { baseCompile as compile, baseParse as parse, run };
//# sourceMappingURL=tiny-sass-compiler.esm-bundler.prod.js.map
